"""Numpy does not incorporate the secant, cosecant functions, as well as their
inverse functions. We do implement it here.

We remind that:
sec = 1/cos
scs = 1/sin
"""

import numpy as np
import numpy.typing as npt


def arccsc(x: float | npt.NDArray[np.float_]) -> float | npt.NDArray[np.float_]:
    return np.arcsin(1 / x)


def arcsec(x: float | npt.NDArray[np.float_]) -> float | npt.NDArray[np.float_]:
    return np.arccos(1 / x)



"""tms.py
Provides analytical functions related to the phase uncertainty during an interferometry
experiment using two-mode squeezed vacuum states as an input state. We denote "nu" the
average population per mode of the TSM. The total average number of particles N is 
therefore N = 2*nu.
The interferometer considered is a generic Mach-Zehnder with a phase difference ϕ
between both arms.
The detectors may have a finite quantum efficiency η.

In order to minimise functions calls, we keep most of the functions in an expanded
analytical form.

abbreviations:
--------------
    - ev: Expectation Value
    - vhd: Variance of the Half Difference of the number of particles at the output
        (i.e. our observable)
    - qe: Quantum efficiency (i.e. the value of η)

general notations:
------------------
    - N: the average  total number of particles in the interferometer. It corresponds to
    the main experimental resource.
    - phi: the phase difference between both arms of the interferometer
    - eta: quantum efficiency of the detectors (between 0 and 1). More generally it may
        characterize any loss process occurring after the last beam splitter of the
        interferometer.
"""

import numpy as np
import numpy.typing as npt
from .trigonometry import arccsc, arcsec

# - OBSERVABLE: Var. of the difference of numbers of particles between both arms
# -------------------------------------------------------------------------------

# |- DETECTORS: perfects
# -----------------------


def ev_vhd_perfect_qe(
    phi: float | npt.NDArray[np.float_],
    N: float | npt.NDArray[np.float_],
) -> float | npt.NDArray[np.float_]:
    """Returns the expectation value of the variance of the half difference of number of
    particles detected at both output ports of the interferometer.
    The detectors are perfect, meaning that their quantum efficiency is equal to 1.
    Since the expectation value of the difference itself is zero, the variance is
    actually equal to the expectation value of the square of the difference.

    Parameters
    ----------
    phi : float | npt.NDArray[np.float_]
        Phase difference between both arms of the interferometer.
    N : float | npt.NDArray[np.float_]
        Average total number of particles. The input state being the two-mode squeezed vacuum
        state with N/2 particles per mode on average.

    Returns
    -------
    float | npt.NDArray[np.float_]
        Expectation value: <(1/4) * (N_output1 - N_output2) ** 2>
    """
    return (N / 2) * (1 + N / 2) * np.sin(phi) ** 2


def ev_vhd_squared_perfect_qe(
    phi: float | npt.NDArray[np.float_],
    N: float | npt.NDArray[np.float_],
) -> float | npt.NDArray[np.float_]:
    """Returns the expectation value of the power four of the half difference of number
    of particles detected at both output ports of the interferometer.
    The detectors are perfect, meaning that their quantum efficiency is equal to 1.

    Parameters
    ----------
    phi : float | npt.NDArray[np.float_]
        Phase difference between both arms of the interferometer.
    N : float | npt.NDArray[np.float_]
        Average total number of particles. The input state being the two-mode squeezed vacuum
        state with N/2 particles per mode on average.

    Returns
    -------
    float | npt.NDArray[np.float_]
        Expectation value: <(1/16) * (N_output1 - N_output2) ** 4>
    """
    return (
        (N / 2)
        * (1 + N / 2)
        * np.sin(phi) ** 2
        * (1 + 9 * N / 2 * (1 + N / 2) * np.sin(2 * phi) ** 2)
    )


def fluctuations_vhd_perfect_qe(
    phi: float | npt.NDArray[np.float_],
    N: float | npt.NDArray[np.float_],
) -> float | npt.NDArray[np.float_]:
    """Returns the quantum fluctuations of the variance of the half difference of number
    of particles detected at both output ports of the interferometer.
    The detectors are perfect, meaning that their quantum efficiency is equal to 1.

    Parameters
    ----------
    phi : float | npt.NDArray[np.float_]
        Phase difference between both arms of the interferometer.
    N : float | npt.NDArray[np.float_]
        Average total number of particles. The input state being the two-mode squeezed vacuum
        state with N/2 particles per mode on average.

    Returns
    -------
    float | npt.NDArray[np.float_]
        Expectation value: <Sqrt[Var( Var{ (1/2) * (N_output1 - N_output2) } )]>
    """
    return np.sqrt(ev_vhd_squared_perfect_qe(phi, N) - ev_vhd_perfect_qe(phi, N) ** 2)


def phase_uncertainty_vhd_perfect_qe(
    phi: float | npt.NDArray[np.float_],
    N: float | npt.NDArray[np.float_],
) -> float | npt.NDArray[np.float_]:
    """Returns the phase uncertainty during an interferometry experiment using
        - two-mode squeezed vacuum states at the input;
        - perfect detectors (quantum efficiency equal to 1);
        - considering the variance of the half difference of particles detected at the
            output as the observable of interest;

    Notice that this resolution depends on the phase difference phi.

    Parameters
    ----------
    phi : float | npt.NDArray[np.float_]
        Phase difference between both arms of the interferometer.
    N : float | npt.NDArray[np.float_]
        Average total number of particles. The input state being the two-mode squeezed vacuum
        state with N/2 particles per mode on average.

    Returns
    -------
    float | npt.NDArray[np.float_]
        Phase uncertainty.
    """
    return np.sqrt(1 + 4 * N * (1 + N / 2) * np.sin(phi) ** 2) / (
        2 * np.cos(phi) * np.sqrt(N / 2 * (1 + N / 2))
    )


# |- DETECTORS: finite quantum efficiency
# ---------------------------------------

# ||- general shape


def ev_vhd_finite_qe(
    phi: float | npt.NDArray[np.float_],
    N: float | npt.NDArray[np.float_],
    eta: float | npt.NDArray[np.float_],
) -> float | npt.NDArray[np.float_]:
    """Returns the expectation value of the variance of the half difference of number of
    particles detected at both output ports of the interferometer.
    The detectors have a finite quantum efficiency eta.
    Since the expectation value of the difference itself is zero, the variance is
    actually equal to the expectation value of the square of the difference.

    Parameters
    ----------
    phi : float | npt.NDArray[np.float_]
        Phase difference between both arms of the interferometer.
    N : float | npt.NDArray[np.float_]
        Average total number of particles. The input state being the two-mode squeezed vacuum
        state with N/2 particles per mode on average.
    eta : float | npt.NDArray[np.float_]
        Quantum efficiency of the detector, must be between 0 and 1.

    Returns
    -------
    float | npt.NDArray[np.float_]
        Expectation value: <(1/4) * (N_output1 - N_output2) ** 2>
    """
    return (
        eta**2 * np.sin(phi) ** 2 * (N / 2) * (1 + N / 2)
        + 0.25 * eta * (1 - eta) * N / 2
    )


def ev_vhd_squared_finite_qe(
    phi: float | npt.NDArray[np.float_],
    N: float | npt.NDArray[np.float_],
    eta: float | npt.NDArray[np.float_],
) -> float | npt.NDArray[np.float_]:
    """Returns the expectation value of the power four of the half difference of number
    of particles detected at both output ports of the interferometer.
    The detectors have a finite quantum efficiency eta.

    Parameters
    ----------
    phi : float | npt.NDArray[np.float_]
        Phase difference between both arms of the interferometer.
    N : float | npt.NDArray[np.float_]
        Average total number of particles. The input state being the two-mode squeezed vacuum
        state with N/2 particles per mode on average.
    eta : float | npt.NDArray[np.float_]
        Quantum efficiency of the detector, must be between 0 and 1.

    Returns
    -------
    float | npt.NDArray[np.float_]
        Expectation value: <(1/16) * (N_output1 - N_output2) ** 4>
    """
    return (eta * N / 128) * (
        8
        + 24 * eta
        + 64 * eta * N
        + 48 * eta**2 * N
        + 72 * eta**2 * N**2
        + 12 * eta**3 * N
        + 36 * eta**3 * N**2
        + 27 * eta**3 * N**3
        - 4
        * eta
        * (2 + N)
        * (4 + 18 * eta * N + 9 * eta**2 * N**2)
        * np.cos(2 * phi)
        + 9 * eta**3 * N * (2 + N) ** 2 * np.cos(4 * phi)
    )


def fluctuations_vhd_finite_qe(
    phi: float | npt.NDArray[np.float_],
    N: float | npt.NDArray[np.float_],
    eta: float | npt.NDArray[np.float_],
) -> float | npt.NDArray[np.float_]:
    """Returns the quantum fluctuations of the variance of the half difference of number
    of particles detected at both output ports of the interferometer.
    The detectors have a finite quantum efficiency eta.

    Parameters
    ----------
    phi : float | npt.NDArray[np.float_]
        Phase difference between both arms of the interferometer.
    N : float | npt.NDArray[np.float_]
        Average total number of particles. The input state being the two-mode squeezed vacuum
        state with N/2 particles per mode on average.
    eta : float | npt.NDArray[np.float_]
        Quantum efficiency of the detector, must be between 0 and 1.

    Returns
    -------
    float | npt.NDArray[np.float_]
        Expectation value: <Sqrt[Var( Var{ (1/2) * (N_output1 - N_output2) } )]>
    """
    return np.sqrt(
        ev_vhd_squared_finite_qe(phi, N, eta) - ev_vhd_finite_qe(phi, N, eta) ** 2
    )


def phase_uncertainty_vhd_finite_qe(
    phi: float | npt.NDArray[np.float_],
    N: float | npt.NDArray[np.float_],
    eta: float | npt.NDArray[np.float_],
) -> float | npt.NDArray[np.float_]:
    """Returns the phase uncertainty during an interferometry experiment using
        - two-mode squeezed vacuum states at the input;
        - detectors with finite quantum efficiency eta;
        - considering the variance of the half difference of particles detected at the
            output as the observable of interest;

    Notice that this resolution depends on the phase difference phi. Here (eta < 1),
    meaning that when (phi = 0), the phase uncertainty goes to infinity.

    Parameters
    ----------
    phi : float | npt.NDArray[np.float_]
        Phase difference between both arms of the interferometer.
    N : float | npt.NDArray[np.float_]
        Average total number of particles. The input state being the two-mode squeezed vacuum
        state with N/2 particles per mode on average.
    eta : float | npt.NDArray[np.float_]
        Quantum efficiency of the detector, must be strictly between 0 and 1.

    Returns
    -------
    float | npt.NDArray[np.float_]
        Phase uncertainty.
    """
    return np.divide(
        np.sqrt(
            N
            * eta
            * (
                1
                + 3 * eta
                + 7 * N * eta
                + 6 * N * eta**2
                + 8 * N**2 * eta**2
                + N * eta**3
                + 4 * N**2 * eta**3
                + 3 * N**3 * eta**3
                - 2
                * (N + 2)
                * eta
                * (1 + 4 * N * eta + 2 * N**2 * eta**2)
                * np.cos(2 * phi)
                + N * (N + 2) ** 2 * eta**3 * np.cos(4 * phi)
            )
        ),
        N * (N + 2) * eta**2 * np.abs(np.sin(2 * phi)),
        # we initialize the output with np.inf, and replace with the proper values where
        # phi!=0. There may be a cleaner way than doing "phi + N + eta" to initialize
        # an array with the correct shape, but the issue is that phi, N, eta can either
        # be scalars, or same-shape arrays....
        out=np.full_like(phi + N + eta, np.inf, dtype=np.float64),
        where=(phi != 0),
    )


# ||- agnostic functions


def ev_vhd(
    phi: float | npt.NDArray[np.float_],
    N: float | npt.NDArray[np.float_],
    eta: float | npt.NDArray[np.float_] = 1,
) -> float | npt.NDArray[np.float_]:
    """Returns the expectation value of the variance of the half difference of number of
    particles detected at both output ports of the interferometer.
    Since the expectation value of the difference itself is zero, the variance is
    actually equal to the expectation value of the square of the difference.

    This function only calls either "ev_vhd_perfect_qe" or "ev_vhd_finite_qe", depending
    on the value of eta that it is set as argument.

    Parameters
    ----------
    phi : float | npt.NDArray[np.float_]
        Phase difference between both arms of the interferometer.
    N : float | npt.NDArray[np.float_]
        Average total number of particles. The input state being the two-mode squeezed vacuum
        state with N/2 particles per mode on average.
    eta : float | npt.NDArray[np.float_], optional
        Quantum efficiency of the detector, must be between 0 and 1, by default 1.

    Returns
    -------
    float | npt.NDArray[np.float_]
        Expectation value: <(1/4) * (N_output1 - N_output2) ** 2>
    """
    return np.where(
        eta == 1,
        ev_vhd_perfect_qe(phi, N),
        ev_vhd_finite_qe(phi, N, eta),
    )


def ev_vhd_squared(
    phi: float | npt.NDArray[np.float_],
    N: float | npt.NDArray[np.float_],
    eta: float | npt.NDArray[np.float_] = 1,
) -> float | npt.NDArray[np.float_]:
    """Returns the expectation value of the power four of the half difference of number
    of particles detected at both output ports of the interferometer.

    This function only calls either "ev_vhd_squared_perfect_qe" or
    "ev_vhd_squared_finite_qe", depending on the value of eta that it is set as
    argument.

    Parameters
    ----------
    phi : float | npt.NDArray[np.float_]
        Phase difference between both arms of the interferometer.
    N : float | npt.NDArray[np.float_]
        Average total number of particles. The input state being the two-mode squeezed vacuum
        state with N/2 particles per mode on average.
    eta : float | npt.NDArray[np.float_], optional
        Quantum efficiency of the detector, must be between 0 and 1, by default 1.

    Returns
    -------
    float | npt.NDArray[np.float_]
        Expectation value: <(1/16) * (N_output1 - N_output2) ** 4>
    """
    return np.where(
        eta == 1,
        ev_vhd_squared_perfect_qe(phi, N),
        ev_vhd_squared_finite_qe(phi, N, eta),
    )


def fluctuations_vhd(
    phi: float | npt.NDArray[np.float_],
    N: float | npt.NDArray[np.float_],
    eta: float | npt.NDArray[np.float_] = 1,
) -> float | npt.NDArray[np.float_]:
    """Returns the quantum fluctuations of the variance of the half difference of number
    of particles detected at both output ports of the interferometer.

    This function only calls either "fluctuations_vhd_perfect_qe" or
    "fluctuations_vhd_finite_qe", depending on the value of eta that it is set as
    argument.

    Parameters
    ----------
    phi : float | npt.NDArray[np.float_]
        Phase difference between both arms of the interferometer.
    N : float | npt.NDArray[np.float_]
        Average total number of particles. The input state being the two-mode squeezed vacuum
        state with N/2 particles per mode on average.
    eta : float | npt.NDArray[np.float_], optional
        Quantum efficiency of the detector, must be between 0 and 1, by default 1.

    Returns
    -------
    float | npt.NDArray[np.float_]
        Expectation value: <Sqrt[Var( Var{ (1/2) * (N_output1 - N_output2) } )]>
    """
    return np.where(
        eta == 1,
        fluctuations_vhd_perfect_qe(phi, N),
        fluctuations_vhd_finite_qe(phi, N, eta),
    )


def phase_uncertainty_vhd(
    phi: float | npt.NDArray[np.float_],
    N: float | npt.NDArray[np.float_],
    eta: float | npt.NDArray[np.float_] = 1,
) -> float | npt.NDArray[np.float_]:
    """Returns the phase uncertainty during an interferometry experiment using
        - two-mode squeezed vacuum states at the input;
        - considering the variance of the half difference of particles detected at the
            output as the observable of interest;

    This function only calls either "phase_uncertainty_vhd_perfect_qe" or
    "phase_uncertainty_vhd_finite_qe", depending on the value of eta that it is set as
    argument.

    Parameters
    ----------
    phi : float | npt.NDArray[np.float_]
        Phase difference between both arms of the interferometer.
    N : float | npt.NDArray[np.float_]
        Average total number of particles. The input state being the two-mode squeezed vacuum
        state with N/2 particles per mode on average.
    eta : float | npt.NDArray[np.float_], optional
        Quantum efficiency of the detector, must be between 0 and 1, by default 1.

    Returns
    -------
    float | npt.NDArray[np.float_]
        Phase uncertainty.
    """
    return np.where(
        eta == 1,
        phase_uncertainty_vhd_perfect_qe(phi, N),
        phase_uncertainty_vhd_finite_qe(phi, N, eta),
    )


# ||- behaviour at the optimal phase


def optimal_phi_vhd(
    N: float | npt.NDArray[np.float_],
    eta: float | npt.NDArray[np.float_] = 1,
) -> float | npt.NDArray[np.float_]:
    """Returns the optimal phase to estimate (minimizing the resolution) during an
    interferometry experiment using a two-mode squeezed vacuum state, detectors with
    finite quantum efficiency and considering the variance of the half difference of
    particles detected at the output as the observable of interest. The detectors have
    a finite quantum efficiency eta.

    Parameters
    ----------
    N : float | npt.NDArray[np.float_]
        Average total number of particles. The input state being the two-mode squeezed vacuum
        state with N/2 particles per mode on average.
    eta : float | npt.NDArray[np.float_], optional
        Quantum efficiency of the detector, must be between 0 and 1, by default 1.

    Returns
    -------
    float | npt.NDArray[np.float_]
        Optimal phase to estimate experimentally.
    """
    return np.where(
        eta == 1,
        0,
        arccsc(
            np.sqrt(
                (1 / ((eta - 1) * (-1 + 5 * eta * N * (eta - 1))))
                * (
                    1
                    + (-1 + 5 * N) * eta
                    - 10 * N * eta**2
                    + 5 * N * eta**3
                    + np.sqrt(
                        (eta - 1)
                        * (
                            -1
                            - 7 * (1 + 2 * N) * eta
                            - N * (52 + 61 * N) * eta**2
                            + N * (30 - 81 * N - 88 * N**2) * eta**3
                            + (85 * N**2 - 40 * N**4) * eta**4
                            + 5 * N**2 * (5 + 16 * N + 8 * N**2) * eta**5
                        )
                    )
                )
            )
        ),
    )


def phase_uncertainty_at_optimal_phi_vhd(
    N: float | npt.NDArray[np.float_],
    eta: float | npt.NDArray[np.float_] = 1,
) -> float | npt.NDArray[np.float_]:
    """Returns the resolution at the optimal phase to estimate during an interferometry
    experiment using a two-mode squeezed vacuum state, detectors with finite quantum
    efficiency eta and considering the variance of the half difference of particles
    detected at the output as the observable of interest.
    This function is therefore just:

    phase_resolution_difference_finite_qe(optimal_phi_vhd(n, eta), n, eta)

    Parameters
    ----------
    N : float | npt.NDArray[np.float_]
        Average total number of particles. The input state being the two-mode squeezed vacuum
        state with N/2 particles per mode on average.
    eta : float | npt.NDArray[np.float_], optional
        Quantum efficiency of the detector, must be between 0 and 1, by default 1.

    Returns
    -------
    float | npt.NDArray[np.float_]
        Optimal resolution.
    """
    nu = N / 2
    return np.where(
        eta == 1,
        1 / np.sqrt(N * (N + 2)),
        1
        / (2 * np.sqrt(2) * eta**2 * nu * (1 + nu))
        * np.sqrt(
            eta
            * nu
            * (
                1
                + eta
                * (
                    3
                    + 2
                    * nu
                    * (7 + eta * (6 + eta + 8 * (2 + eta) * nu + 12 * eta * nu**2))
                )
                + 4
                * eta
                * (1 + nu)
                * (1 + 8 * eta * nu * (1 + eta * nu))
                * np.cos(
                    2
                    * arcsec(
                        np.sqrt(
                            (
                                1
                                + eta * (-1 + 10 * (-1 + eta) ** 2 * nu)
                                + np.sqrt(
                                    (-1 + eta)
                                    * (-1 + 10 * (-1 + eta) * eta * nu)
                                    * (
                                        1
                                        + eta
                                        * (
                                            7
                                            + 2
                                            * nu
                                            * (
                                                9
                                                + eta
                                                * (
                                                    22
                                                    + 32 * nu
                                                    + eta * (5 + 32 * nu * (1 + nu))
                                                )
                                            )
                                        )
                                    )
                                )
                            )
                            / (1 - eta + 10 * (-1 + eta) ** 2 * eta * nu)
                        )
                    )
                )
                + 8
                * eta**3
                * nu
                * (1 + nu) ** 2
                * np.cos(
                    4
                    * arcsec(
                        np.sqrt(
                            (
                                1
                                + eta * (-1 + 10 * (-1 + eta) ** 2 * nu)
                                + np.sqrt(
                                    (-1 + eta)
                                    * (-1 + 10 * (-1 + eta) * eta * nu)
                                    * (
                                        1
                                        + eta
                                        * (
                                            7
                                            + 2
                                            * nu
                                            * (
                                                9
                                                + eta
                                                * (
                                                    22
                                                    + 32 * nu
                                                    + eta * (5 + 32 * nu * (1 + nu))
                                                )
                                            )
                                        )
                                    )
                                )
                            )
                            / (1 - eta + 10 * (-1 + eta) ** 2 * eta * nu)
                        )
                    )
                )
            )
        )
        / np.sin(
            2
            * arcsec(
                np.sqrt(
                    (
                        1
                        + eta * (-1 + 10 * (-1 + eta) ** 2 * nu)
                        + np.sqrt(
                            (-1 + eta)
                            * (-1 + 10 * (-1 + eta) * eta * nu)
                            * (
                                1
                                + eta
                                * (
                                    7
                                    + 2
                                    * nu
                                    * (
                                        9
                                        + eta
                                        * (
                                            22
                                            + 32 * nu
                                            + eta * (5 + 32 * nu * (1 + nu))
                                        )
                                    )
                                )
                            )
                        )
                    )
                    / (1 - eta + 10 * (-1 + eta) ** 2 * eta * nu)
                )
            )
        ),
    )


def asymptotic_ratio_phase_uncertainty_to_SQL_at_optimal_phi_vhd(
    eta: float | npt.NDArray[np.float_],
) -> float | npt.NDArray[np.float_]:
    """Returns the asymptotic limit (as the number of particles goes to infinity) of the ratio between:
        - the phase uncertainty at the optimal phase and considering the variance of the half difference
        of particles detected at the output as the observable of interest
        - the SQL 1/sqrt(eta N).

    It only depends on the quantum efficiency of the detectors.

    Parameters
    ----------
    eta : float | npt.NDArray[np.float_]
        Quantum efficiency of the detector, must be between 0 and 1.

    Returns
    -------
    float | npt.NDArray[np.float_]
        Optimal phase uncertainty to SQL ratio in the asymptotic limit of N.
    """
    return ((2 / 5) ** (1 / 4) * np.sqrt(5 + 2 * np.sqrt(10))) * np.sqrt(1 - eta)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['qsipy']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.25.0,<2.0.0']

setup_kwargs = {
    'name': 'qsipy',
    'version': '0.1.1',
    'description': 'Quantum State Interferometry with PYthon.',
    'long_description': '# Python package about interferometry using two-mode quantum states\n\n## This package provides functions that return various physical quantities related to interferometry experiments using twin Fock states $|n,n\\rangle$ or two-mode squeezed vacuum states\n\n### Functions contained in this package are derived in [this article](link)\n* DOI: *paper published soon*\n---\n\n## Quick context explanation\n\nThe scheme of the experiment that we aer considering is represented with the figure below:\n![scheme_experiment](images/schematic_exp.png)\n* **twin Fock** or **two-mode squeezed** states are used as *input* in a Mach-Zehnder interferometer ;\n* the detection suffers from a **non-unit quantum efficiency** $\\eta$ ;\n* experimentalists are interested in the measurement of the phase difference $\\phi$ between the two arms ;\n* the observable $\\hat{O}$ that one use to perform a measurement is the **square** of the *half difference* of particles detected at both output ports $\\hat{c}_1$ and $\\hat{c}_2$, whose expectation value is actually the **variance** of the *half difference* of particles detected[^1]:\n    $$\\hat{O} = \\frac{1}{4} \\left( \\hat{N}_{c_2} - \\hat{N}_{c_1} \\right)^2$$\n* detailed explanation about the derivation of the formulae are given in the [supplemental material](link) of the article ;\n\n[^1]: indeed, $\\langle \\hat{N}_{c_2} - \\hat{N}_{c_1} \\rangle = 0$, and therefore  $4 \\times \\mathrm{Var} \\left[ \\hat{O} \\right] =  \\left\\langle \\left(\\hat{N}_{c_2} - \\hat{N}_{c_1}\\right)^2 \\right\\rangle - \\left\\langle \\hat{N}_{c_2} - \\hat{N}_{c_1} \\right\\rangle^2 = \\left\\langle \\left(\\hat{N}_{c_2} - \\hat{N}_{c_1}\\right)^2 \\right\\rangle$\n---\n\n## Installation\n\n#### Method 1: download from PyPI (recommended for users)\nThis package is published in the PyPI repository, it can be added to any Python environment with\nthe oneliner:\n`pip install qsipy`\n\nOr with any other Python packaging and dependency manager, such as poetry:\n`poetry add qsipy`\n\n#### Method 2: cloning from the source (for contributers)\n\nYou may also clone this project and use it right away. Since it is only Python scripts\nno compilation is needed. The functions of interest are defined in two separate modules:\n* `src/qsipy/tfs.py` $\\Rightarrow$ for the functions related to the *twin Fock states*\n* `src/qsipy/tms.py` $\\Rightarrow$ for the functions related to the *two-mode squeezed vacuum states* \n---\n\n## Examples\n\n#### 1. Phase uncertainty $\\Delta \\phi$ as a function of the measured phase difference\n\nThe following figure shows the **ratio** between the **phase sensitivity** achieved with those two quantum states **and the shot-noise**. The quantum efficiency is set to $\\eta = 0.95$. It shows in particular that interferometry [below the standard quantum limit (SQL)](https://arxiv.org/abs/1405.7703) can be obtained in specific ranges of phase differences $\\phi$.\n\n![tf_vs_tms](images/FIG_tf_vs_tms/qsipy_tf_vs_tms.png)\n\nThe source code generating this figure [can be found here](images/FIG_tf_vs_tms/qsipy_tf_vs_tms.py).\n\n#### 2. Maps of the optimal $\\phi_0$ to measure, and sub-shot-noise domains\n\nThe example 1. exhibited the fact that (for given quantum efficiency $\\eta$ and number of particles $N$) there is an optimal phase difference $\\phi_0$ to measure, in order to minimize the measurement uncertainty $\\Delta \\phi$. One can plot the maps of those optimal phases in the $(\\eta,N)$ plane, and identify the regions where sub-shot-noise interferometry may be performed.\n\n* `TF` stands for *twin Fock* state\n* `TMS` stands for *two-mode squeezed vacuum* state\n\n![optimal_phi](images/FIG_optimal_phi/qsipy_optimal_phi.png)\n\nThe source code generating this figure [can be found here](images/FIG_optimal_phi/qsipy_optimal_phi.py).\n\n#### 3. Asymptotic behaviour of the phase uncertainty\n\nKnowing the optimal phase $\\phi_0$ to estimate, one can study the asymptotic behaviour of $\\Delta \\phi_0 = \\left.\\Delta \\phi\\right|_{\\phi=\\phi_0}$. One can therefore check that as soon as $\\eta \\neq 1$, $\\Delta \\phi_0$ has a $\\mathcal{O}(N^{-1/2})$ scaling, therefore overtaking the standard quantum limit only by a constant factor.\n\n---\n\n## User guide\n\nCurrently the package is organized in two main modules `tfs` and `tms`, containing essentially the same functions, but considering either *twin Fock* or *two-mode squeezed* states at the input.\n\n**Important abbreviations:** for the sake of compactness we use the following abbreviation in the names of the functions provided by `qsipy`:\n* `ev` stands for *"expectation value"*\n* `vhd` stands for *"variance of the half difference"*\n* `qe` stands for *"quantum efficiency"* (i.e. the value of $\\eta$)\n\nHere is the listing of the useful functions that one can use, calling `tfs.`*myfunction*\nor `tfs.`*myfunction*:\n\n#### Most important functions\n\n```python\ndef phase_uncertainty_vhd(\n    phi: float | npt.NDArray[np.float_],\n    N: float | npt.NDArray[np.float_],\n    eta: float | npt.NDArray[np.float_] = 1,\n) -> float | npt.NDArray[np.float_]:\n    """Returns the phase uncertainty during an interferometry experiment using\n        - twin-Fock or two-mode squeezed states at the input (depending from which module\n            the function has been called);\n        - considering the variance of the half difference of particles detected at the\n            output as the observable of interest;\n\n    This function only calls either "phase_uncertainty_vhd_perfect_qe" or\n    "phase_uncertainty_vhd_finite_qe", depending on the value of eta that it is set as\n    argument.\n\n    Parameters\n    ----------\n    phi : float | npt.NDArray[np.float_]\n        Phase difference between both arms of the interferometer.\n    N : float | npt.NDArray[np.float_]\n        Total number of particles.\n    eta : float | npt.NDArray[np.float_], optional\n        Quantum efficiency of the detector, must be between 0 and 1, by default 1.\n\n    Returns\n    -------\n    float | npt.NDArray[np.float_]\n        Phase uncertainty.\n    """\n```\n\n```python\ndef optimal_phi_vhd(\n    N: float | npt.NDArray[np.float_],\n    eta: float | npt.NDArray[np.float_] = 1,\n) -> float | npt.NDArray[np.float_]:\n    """Returns the optimal phase to estimate (minimizing the resolution) during an\n    interferometry experiment using detectors with finite quantum efficiency and\n    considering the variance of the half difference of particles detected at the output\n    as the observable of interest. The detectors have a finite quantum efficiency eta.\n\n    Parameters\n    ----------\n    N : float | npt.NDArray[np.float_]\n        Total number of particles.\n    eta : float | npt.NDArray[np.float_], optional\n        Quantum efficiency of the detector, must be between 0 and 1, by default 1.\n\n    Returns\n    -------\n    float | npt.NDArray[np.float_]\n        Optimal phase to estimate experimentally.\n    """\n```\n\n```python\ndef phase_uncertainty_at_optimal_phi_vhd(\n    N: float | npt.NDArray[np.float_],\n    eta: float | npt.NDArray[np.float_] = 1,\n) -> float | npt.NDArray[np.float_]:\n    """Returns the resolution at the optimal phase to estimate during an interferometry\n    experiment using detectors with finite quantum efficiency eta and considering the\n    variance of the difference of particles at the output as the observable of interest.\n    This function is therefore just:\n\n    phase_resolution_difference_finite_qe(optimal_phi_vhd(n, eta), n, eta)\n\n    Parameters\n    ----------\n    N : float | npt.NDArray[np.float_]\n        Total number of particles.\n    eta : float | npt.NDArray[np.float_], optional\n        Quantum efficiency of the detector, must be between 0 and 1, by default 1.\n\n    Returns\n    -------\n    float | npt.NDArray[np.float_]\n        Optimal resolution.\n    """\n```\n\n#### Functions with more specific usage\n\n```python\ndef ev_vhd(\n    phi: float | npt.NDArray[np.float_],\n    N: float | npt.NDArray[np.float_],\n    eta: float | npt.NDArray[np.float_] = 1,\n) -> float | npt.NDArray[np.float_]:\n    """Returns the expectation value of the variance of the half difference of number of\n    particles detected at both output ports of the interferometer.\n    Since the expectation value of the difference itself is zero, the variance is\n    actually equal to the expectation value of the square of the difference.\n\n    This function only calls either "ev_vhd_perfect_qe" or "ev_vhd_finite_qe", depending\n    on the value of eta that it is set as argument.\n\n    Parameters\n    ----------\n    phi : float | npt.NDArray[np.float_]\n        Phase difference between both arms of the interferometer.\n    N : float | npt.NDArray[np.float_]\n        Average total number of particles in the interferometer.\n    eta : float | npt.NDArray[np.float_], optional\n        Quantum efficiency of the detector, must be between 0 and 1, by default 1.\n\n    Returns\n    -------\n    float | npt.NDArray[np.float_]\n        Expectation value: <(1/4) * (N_output1 - N_output2) ** 2>\n    """\n```\n\n```python\ndef ev_vhd_squared(\n    phi: float | npt.NDArray[np.float_],\n    N: float | npt.NDArray[np.float_],\n    eta: float | npt.NDArray[np.float_] = 1,\n) -> float | npt.NDArray[np.float_]:\n    """Returns the expectation value of the power four of the half difference of number\n    of particles detected at both output ports of the interferometer.\n\n    This function only calls either "ev_vhd_squared_perfect_qe" or\n    "ev_vhd_squared_finite_qe", depending on the value of eta that it is set as\n    argument.\n\n    Parameters\n    ----------\n    phi : float | npt.NDArray[np.float_]\n        Phase difference between both arms of the interferometer.\n    N : float | npt.NDArray[np.float_]\n        Total number of particles.\n    eta : float | npt.NDArray[np.float_], optional\n        Quantum efficiency of the detector, must be between 0 and 1, by default 1.\n\n    Returns\n    -------\n    float | npt.NDArray[np.float_]\n        Expectation value: <(1/16) * (N_output1 - N_output2) ** 4>\n    """\n```\n\n```python\ndef fluctuations_vhd(\n    phi: float | npt.NDArray[np.float_],\n    N: float | npt.NDArray[np.float_],\n    eta: float | npt.NDArray[np.float_] = 1,\n) -> float | npt.NDArray[np.float_]:\n    """Returns the quantum fluctuations of the variance of the half difference of number\n    of particles detected at both output ports of the interferometer.\n\n    This function only calls either "fluctuations_vhd_perfect_qe" or\n    "fluctuations_vhd_finite_qe", depending on the value of eta that it is set as\n    argument.\n\n    Parameters\n    ----------\n    phi : float | npt.NDArray[np.float_]\n        Phase difference between both arms of the interferometer.\n    N : float | npt.NDArray[np.float_]\n        Total number of particles.\n    eta : float | npt.NDArray[np.float_], optional\n        Quantum efficiency of the detector, must be between 0 and 1, by default 1.\n\n    Returns\n    -------\n    float | npt.NDArray[np.float_]\n        Expectation value: <Sqrt[Var( Var{ (1/2) * (N_output1 - N_output2) } )]>\n    """\n```\n\n```python\ndef asymptotic_ratio_phase_uncertainty_to_SQL_at_optimal_phi_vhd(\n    eta: float | npt.NDArray[np.float_],\n) -> float | npt.NDArray[np.float_]:\n    """Returns the asymptotic limit (as the number of particles goes to infinity) of the ratio between:\n        - the phase uncertainty at the optimal phase and considering the variance of the half difference\n        of particles detected at the output as the observable of interest\n        - the SQL 1/sqrt(eta N).\n\n    It only depends on the quantum efficiency of the detectors.\n\n    Parameters\n    ----------\n    eta : float | npt.NDArray[np.float_]\n        Quantum efficiency of the detector, must be between 0 and 1.\n\n    Returns\n    -------\n    float | npt.NDArray[np.float_]\n        Optimal phase uncertainty to SQL ratio in the asymptotic limit of N.\n    """\n```\n',
    'author': 'Quentin Marolleau',
    'author_email': 'quentin.marolleau@institutoptique.fr',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
